﻿/*! ------------------------------------------------------------------------
//                                   Fusion
//  ------------------------------------------------------------------------
//
//                       Copyright 2014 Nicholas Gaulin
//
//       Licensed under the Apache License, Version 2.0 (the "License");
//      you may not use this file except in compliance with the License.
//                   You may obtain a copy of the License at
//
//                 http://www.apache.org/licenses/LICENSE-2.0
//
//     Unless required by applicable law or agreed to in writing, software
//      distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//     See the License for the specific language governing permissions and
//                       limitations under the License.
*/

// ########## LEXER ##########
var Lexer   = function($source, $language)
{
    // If source code was provided, set it
    if ($source)
        this.source = $source;

    // If a language wasn't provided, return
    if (!$language)
        return;

    // Set the language and create the scope chain
    this.language = $language;
    this.chain    = $language == 'fhtml' ?
                    [new Scope('')] :
                    $language == 'fjs'
                 || $language == 'fcss' ?
                    [] :
                    null;
},
  __Lexer__ = Lexer.prototype = {};

// --- HELPERS ---
var $_lexer_clone      = function($v, $i, $a)
{
    // Return a clone of the value
    return $v.clone();
};
var $_lexer_comment    = function($state)
{
    // Return true if the state is a comment
    return ($state == 'JavaScriptBlockComment'
         || $state == 'JavaScriptLineComment'
         || $state == 'HTMLCommentOpen'
         || $state == 'HTMLCommentText'
         || $state == 'HTMLCommentClose'
         || $state == 'HTMLBogusCommentOpen'
         || $state == 'HTMLBogusCommentText'
         || $state == 'HTMLBogusCommentClose'
         || $state == 'CSSComment');
};
var $_lexer_count      = function($scope, $character)
{
    // If the character is either an opening or closing brace, adjust the braces count of the scope
    if ($character == '{' || $character == '}')
        $scope.braces += $character == '{' ?
                         1 :
                         $scope.braces ?
                         -1 :
                         0;
    // If the character is either an opening or closing parentheses, adjust the parentheses count of the scope
    else if ($character == '(' || $character == ')')
        $scope.parentheses += $character == '(' ?
                              1 :
                              $scope.parentheses ?
                              -1 :
                              0;
    // If the character is either an opening or closing bracket, adjust the brackets count of the scope
    else if ($character == '[' || $character == ']')
        $scope.brackets += $character == '[' ?
                           1 :
                           $scope.brackets ?
                           -1 :
                           0;
    // If the character is either an opening or closing tag, adjust the tags count of the scope
    else if ($character == '<' || $character == '>')
        $scope.tags += $character == '<' ?
                       1 :
                       $scope.tags ?
                       -1 :
                       0;
};
var $_lexer_regexp     = function($token, $expression)
{
    // If no token was provided, return true
    if (!$token)
        return true;

    // Get the token type
    var $type = $token.type;

    // If the token is a punctuator
    if ($type == 'JavaScriptPunctuator')
    {
        // Get the token text
        var $text = $token.text();

        // Return true if the punctuator cannot preceed a division operator
        return (($text != ')' || $expression && ($expression == 'for()'
                                              || $expression == 'if()'
                                            //|| $expression == 'let()'
                                              || $expression == 'while()'
                                              || $expression == 'with()'))
             && ($text != '}' || !$expression || $expression != '{}')
             &&  $text != ']'
             &&  $text != '++'
             &&  $text != '--');
    }
    // If the token is a reserved word
    else if ($type == 'JavaScriptReservedWord')
    {
        // Get the token text
        var $text = $token.text();

        // Return true if the reserved word can preceed an expression
        return ($text == 'case'
             || $text == 'delete'
             || $text == 'do'
             || $text == 'else'
             || $text == 'extends'
             || $text == 'in'
             || $text == 'instanceof'
             || $text == 'new'
             || $text == 'return'
             || $text == 'throw'
             || $text == 'typeof'
             || $text == 'void'
             || $text == 'yield');
    }
    // If the token is a template string, return true if the template string is a substitution head or middle
    else if ($type == 'JavaScriptTemplateString')
        return $token.source.substr($token.end - 1, 1) != '`';

    // Return true if the token is either a template string head, middle, or substitution open
    return ($type == 'FusionAttributeTemplateStringHead'
         || $type == 'FusionAttributeTemplateStringMiddle'
         || $type == 'FusionSubstitutionOpen'
         || $type == 'FusionObjectSubstitutionOpen'
         || $type == 'FusionSelectorSubstitutionOpen'
         || $type == 'FusionStyleSubstitutionOpen');
};
var $_lexer_text       = function($state)
{
    // Return true if the state is plain text
    return ($state == 'JavaScriptInvalidCharacter'
         || $state == 'HTMLText'
         || $state == 'HTMLEndTagText');
};
var $_lexer_void       = function($tag)
{
    // Return true if the tag is a void HTML element
    return ($tag == 'area'
         || $tag == 'base'
         || $tag == 'br'
         || $tag == 'col'
         || $tag == 'embed'
         || $tag == 'hr'
         || $tag == 'img'
         || $tag == 'input'
         || $tag == 'keygen'
         || $tag == 'link'
         || $tag == 'menuitem'
         || $tag == 'meta'
         || $tag == 'param'
         || $tag == 'source'
         || $tag == 'track'
         || $tag == 'wbr');
};
var $_lexer_whitespace = function($state)
{
    // Return true if the state is whitespace
    return ($state == 'JavaScriptWhitespace'
         || $state == 'HTMLStartTagWhitespace'
         || $state == 'HTMLEndTagWhitespace'
         || $state == 'HTMLDOCTYPEWhitespace'
         || $state == 'CSSWhitespace');
};

// --- PROTOTYPE ---
__Lexer__.chain      = null;
__Lexer__.expression = '';
__Lexer__.language   = '';
__Lexer__.position   = 0;
__Lexer__.source     = '';
__Lexer__.state      = '';
__Lexer__.token      = null;

$_data(__Lexer__, 'clone',  function()
{
    // Create a context-free lexer
    var $lexer = new Lexer();

    // Copy the lexer parameters
    $lexer.chain      = this.chain ?
                        this.chain.map($_lexer_clone) :
                        null;
    $lexer.expression = this.expression;
    $lexer.language   = this.language;
    $lexer.position   = this.position;
    $lexer.source     = this.source;
    $lexer.state      = this.state;
    $lexer.token      = this.token ?
                        this.token.clone() :
                        null;

    // Return the lexer
    return $lexer;
}, true);
$_data(__Lexer__, 'equals', function($lexer)
{
    // Get the scope chains
    var $chainLexer = $lexer.chain,
        $chainThis  = this.chain;

    // If the lexers don't have matching scope chain lengths, return false
    if (!($chainThis && $chainThis.length) != !($chainLexer && $chainLexer.length))
        return false;

    // If the lexers have scope chains, return false if they are not equal
    if ($chainThis)
        for (var $i = 0, $j = $chainThis.length; $i < $j; $i++)
            if (!$chainThis[$i].equals($chainLexer[$i]))
                return false;

    // If the lexers don't have matching previous tokens, return false
    if (!this.token != !$lexer.token || this.token && !this.token.equals($lexer.token))
        return false;

    // Return true if the lexer has the same previous expression, language, and state
    return (this.expression === $lexer.expression
         && this.language   === $lexer.language
         && this.state      === $lexer.state);
}, true);
$_data(__Lexer__, 'hack',   function()
{
    // If a scope chain already exists, return false
    if (this.chain)
        return false;

    // Get the language
    var $language = this.language;

    // Create the scope chain
    this.chain = $language == 'html'
              || $language == 'fhtml' ?
                 [new Scope('')] :
                 $language == 'js'
              || $language == 'fjs'
              || $language == 'fcss' ?
                 [] :
                 null;

    // Return true if a scope chain was created
    return !!this.chain;
}, true);
$_data(__Lexer__, 'next',   function()
{
    // Get the current position, state, and scope
    var $source     = this.source,
        $length     = $source.length,
        $start      = this.position,
        $token      = this.token,
        $expression = this.expression,
        $language   = this.language,
        $position   = this.position,
        $state      = this.state || $language,
        $fusion     = $language == 'fjs'
                   || $language == 'fhtml'
                   || $language == 'fcss',
        $chain      = this.chain,
        $scopes     = $chain ?
                      $chain.length :
                      0,
        $scope      = $scopes ?
                      $chain[$scopes - 1] :
                      null;

    // If the position is invalid, return null
    if ($position < 0 || $position >= $length)
        return null;

    // Create the context parameters
    var $append  = '',
        $end     = false,
        $pop     = false,
        $push    = null,
        $replace = '';

    switch ($state)
    {
        // ---------- JavaScript ----------
        case 'js':
        case 'fjs':
        case 'JavaScriptIdentifier':
        case 'JavaScriptReservedWord':
        case 'JavaScriptPunctuator':
        case 'JavaScriptNumber':
        case 'JavaScriptDoubleQuotedString':
        case 'JavaScriptSingleQuotedString':
        case 'JavaScriptTemplateString':
        case 'JavaScriptRegExp':
        case 'JavaScriptBlockComment':
        case 'JavaScriptLineComment':
        case 'JavaScriptInvalidCharacter':
        case 'JavaScriptWhitespace':
        case 'FusionStartTagOpen':
        case 'FusionEndTagOpen':
        case 'FusionDirectiveTagOpen':
        case 'FusionProperty':
        case 'FusionAttributeTemplateStringHead':
        case 'FusionAttributeTemplateStringMiddle':
        case 'FusionSubstitutionOpen':
        case 'FusionObjectSubstitutionOpen':
        case 'FusionSelectorSubstitutionOpen':
        case 'FusionStyleSubstitutionOpen':

            // Get the current character from the source
            var $current = $source[$position];

            // NAMES AND KEYWORDS (11.6)
            // LITERALS (11.8)
            if ($_letter($current) || $current == '_'
                                   || $current == '$'
                                   || $current == '\\'
                                   || $current == '@' && $fusion && $token && ($token.type == 'JavaScriptIdentifier'
                                                                           || ($token.type == 'JavaScriptPunctuator' || $token.type == 'CSSPunctuator') && ($source[$token.start] == ')'
                                                                                                                                                         || $source[$token.start] == ']')))
            {
                // If the current character isn't an escape sequence, consume the current character
                if ($current != '\\')
                    $position++;

                // Set the state
                $state = $current == '@' ?
                         'FusionProperty' :
                         'JavaScriptIdentifier';

                do
                {
                    // Get the current character from the source
                    $current = $source[$position];

                    // If the current character is a valid identifier character
                    if ($_alphanumeric($current) || $current == '_'
                                                 || $current == '$'
                                                 || $current == '\u200C'
                                                 || $current == '\u200D')
                        // Consume the current character
                        continue;

                    // If the current character isn't an escape sequence, break (and don't consume it)
                    if ($current != '\\')
                        break;

                    // Get the next character from the source
                    var $peek = $source[$position + 1];

                    // If the current character isn't an escape sequence, break
                    if ($peek != 'u')
                        break;

                    // Get the next character from the source
                    $peek = $source[$position + 2];

                    // If the next character opens an ECMAScript 6 unicode escape sequence
                    if ($peek == '{')
                    {
                        // Create the character counter
                        var $characters = 0;

                        // Count the next hexadecimal characters
                        while ($_hexadecimal($source[$position + $characters + 3]))
                            $characters++;

                        // If there are any hexadecimal characters and the escape sequence is properly closed, consume the ECMAScript 6 unicode escape sequence
                        if ($characters && $source[$position + $characters + 3] == '}')
                            $position += $characters + 3;
                        else
                            break;
                    }
                    // If the following four characters are hexadecimal
                    else if ($_hexadecimal($peek) && $_hexadecimal($source[$position + 3])
                                                  && $_hexadecimal($source[$position + 4])
                                                  && $_hexadecimal($source[$position + 5]))
                        // Consume the unicode escape sequence
                        $position += 5;
                    else
                        break;
                }
                // Continue if the incremented position doesn't exceed the length
                while (++$position < $length);

                // If no characters were consumed
                if ($position == $start)
                {
                    // Consume the current character and set the state
                    $position++;
                    $state = 'JavaScriptInvalidCharacter';
                }
                // IDENTIFIER NAMES (11.6.1)
                else if ((!$token || $token.type != 'JavaScriptPunctuator' || $token.text() != '.') && (!$fusion || $state != 'FusionProperty'))
                {
                    // Get the identifier from the source
                    var $identifier = $source.substring($start, $position);

                    // If the previous token was a fusion markup tag and the current identifier opens a fusion context
                    if ($fusion && $token && ($token.type == 'FusionStartTagOpen' && ($identifier == 'break'
                                                                                   || $identifier == 'case'
                                                                                   || $identifier == 'continue'
                                                                                   || $identifier == 'default'
                                                                                   || $identifier == 'do'
                                                                                   || $identifier == 'else'
                                                                                   || $identifier == 'for'
                                                                                   || $identifier == 'if'
                                                                                   || $identifier == 'switch'
                                                                                   || $identifier == 'while')
                                           || $token.type == 'FusionEndTagOpen'   &&  $identifier == 'while'))
                    {
                        // Set the state and push a fusion context into the scope chain
                        $state = 'JavaScriptReservedWord';
                        $push  = new Scope($token.type == 'FusionEndTagOpen' ?
                                           '</@' + $identifier :
                                           '<@' + $identifier);
                    }
                    // RESERVED WORDS (11.6.2)
                    // NULL LITERALS (11.8.1)
                    // BOOLEAN LITERALS (11.8.2)
                    else if ($identifier == 'null'
                          || $identifier == 'false'
                          || $identifier == 'true'
                          || $identifier == 'break'
                          || $identifier == 'case'
                          || $identifier == 'catch'
                          || $identifier == 'class'
                          || $identifier == 'const'
                          || $identifier == 'continue'
                          || $identifier == 'debugger'
                          || $identifier == 'default'
                          || $identifier == 'delete'
                          || $identifier == 'do'
                          || $identifier == 'else'
                          || $identifier == 'enum'
                          || $identifier == 'export'
                          || $identifier == 'extends'
                          || $identifier == 'finally'
                          || $identifier == 'for'
                          || $identifier == 'function'
                          || $identifier == 'if'
                          || $identifier == 'implements'
                          || $identifier == 'import'
                          || $identifier == 'in'
                          || $identifier == 'instanceof'
                          || $identifier == 'interface'
                          || $identifier == 'let'
                          || $identifier == 'new'
                          || $identifier == 'package'
                          || $identifier == 'private'
                          || $identifier == 'protected'
                          || $identifier == 'public'
                          || $identifier == 'return'
                          || $identifier == 'static'
                          || $identifier == 'super'
                          || $identifier == 'switch'
                          || $identifier == 'this'
                          || $identifier == 'throw'
                          || $identifier == 'try'
                          || $identifier == 'typeof'
                          || $identifier == 'var'
                          || $identifier == 'void'
                          || $identifier == 'while'
                          || $identifier == 'with'
                          || $identifier == 'yield')
                    {
                        // Get the previous token text if it's a reserved word
                        var $text = $token && $token.type == 'JavaScriptReservedWord' ?
                                    $token.text() :
                                    null;

                        // If the previous token is neither the get nor set reserved word and the current context isn't an object key context, set the state
                        if ((!$text || $text != 'get' && $text != 'set') && (!$scope || $scope.state != '{'))
                            $state = 'JavaScriptReservedWord';

                        // If the previous token is the `else` reserved word and the identifier can be appended to the state of the current context
                        if ($fusion && $text == 'else' && $scope && $scope.state == '<@else' && ($identifier == 'do'
                                                                                              || $identifier == 'for'
                                                                                              || $identifier == 'if'
                                                                                              || $identifier == 'switch'
                                                                                              || $identifier == 'while'))
                            // Append the identifier to the state of the current context
                            $append = ' ' + $identifier;
                    }
                    // If the identifier is either `get` or `set` and the current context is an object key
                    else if (($identifier == 'get'
                           || $identifier == 'set') && $scope && $scope.state == '{')
                    {
                        // Get the next token and its type
                        var $peek = this.peek($position),
                            $type = $peek ?
                                    $peek.type :
                                    null;

                        // If the next token is either an identifier or reserved word
                        if ($type == 'JavaScriptIdentifier' || $type == 'JavaScriptReservedWord')
                        {
                            // Get the next token and its type
                            $peek = this.peek($peek.end),
                            $type = $peek ?
                                    $peek.type :
                                    null;

                            // If the next token is an opening parentheses
                            if ($type == 'JavaScriptPunctuator' && $source[$peek.start] == '(')
                            {
                                // Set the state and replace the state of the current context with an object value context
                                $state   = 'JavaScriptReservedWord';
                                $replace = '{:';
                            }
                        }
                    }
                }
            }
            // WHITE SPACE (11.2)
            // LINE TERMINATORS (11.3)
            else if ($_whitespace($current) || $_newline($current))
            {
                // Consume the current character and set the state
                $position++;
                $state = 'JavaScriptWhitespace';

                do
                {
                    // Get the current character from the source
                    $current = $source[$position];

                    // If the current character is neither whitespace nor a newline, break (and don't consume it)
                    if (!$_whitespace($current) && !$_newline($current))
                        break;
                }
                // Continue if the incremented position doesn't exceed the length
                while (++$position < $length);
            }
            // STRING LITERALS (11.8.4)
            // TEMPLATE LITERALS (11.8.6)
            else if ($current == '"'
                  || $current == "'"
                  || $current == '`' && $chain
                  || $current == '}' && $scope && $scope.braces == 0 && $_endsWith($scope.state, '`${'))
            {
                // Check if the template string is an attribute and get the break character
                var $attribute = $current == '}' && $scope.state == '=`${',
                    $break     = $current != '}' ?
                                 $current :
                                 '`';

                // Consume the current character and set the state
                $position++;
                $state = $attribute ?
                         'FusionAttributeTemplateStringTail' :
                         $break == '"' ?
                         'JavaScriptDoubleQuotedString' :
                         $break == "'" ?
                         'JavaScriptSingleQuotedString' :
                         'JavaScriptTemplateString';

                // If the template string is either a middle or tail substitution
                if ($current == '}')
                {
                    // Append the characters to the state of the current context and pop it from the scope chain
                    $append = $current + $break;
                    $pop    = true;
                }

                do
                {
                    // Get the current character from the source
                    $current = $source[$position];

                    // If the current character is an escape sequence
                    if ($current == '\\')
                    {
                        // Consume the escape character and if there isn't another character, break
                        if (++$position >= $length)
                            break;

                        // If the current character is a carriage return and the next character is a line break, consume the carriage return
                        if ($source[$position] == '\r' && $source[$position + 1] == '\n')
                            $position++;
                    }
                    // If the current character is the break character
                    else if ($current == $break)
                    {
                        // Consume the current character
                        $position++;

                        break;
                    }
                    // If the state is a template string
                    else if ($break == '`')
                    {
                        // If the current character opens a template substitution
                        if ($current == '$' && $source[$position + 1] == '{')
                        {
                            // Consume the `${` characters and push a substitution context into the scope chain
                            $position += 2;
                            $push      = new Scope($pop ?
                                                   $scope.state :
                                                   '`${');

                            // If the template string is an attribute, set the state
                            if ($attribute)
                                $state = 'FusionAttributeTemplateStringMiddle';

                            break;
                        }
                    }
                    // LINE TERMINATORS (11.3)
                    else if ($_newline($current))
                        break;
                }
                // Continue if the incremented position doesn't exceed the length
                while (++$position < $length);
            }
            // If the current character could be a comment, regular expression literal, or division punctuator (then prepare for some ambiguity)
            else if ($current == '/')
            {
                // Get the next character from the source
                var $peek = $source[$position + 1];

                // COMMENTS (11.4)
                if ($peek == '*')
                {
                    // Consume both characters and set the state
                    $position += 2;
                    $state     = 'JavaScriptBlockComment';

                    do
                    {
                        // If the current characters close the comment
                        if ($source[$position] == '*' && $source[$position + 1] == '/')
                        {
                            // Consume the `*/` characters
                            $position += 2;

                            break;
                        }
                    }
                    // Continue if the incremented position doesn't exceed the length
                    while (++$position < $length);
                }
                // COMMENTS (11.4)
                else if ($peek == '/')
                {
                    // Consume both characters and set the state
                    $position += 2;
                    $state     = 'JavaScriptLineComment';

                    // Consume any other characters until a newline is reached
                    while (!$_newline($source[$position]))
                        $position++;
                }
                // If the current character is the self closing of a fusion start tag
                else if ($fusion && $peek == '>' && $scope && $_startsWith($scope.state, '<@') && ($scope.state == '<@break'
                                                                                                || $scope.state == '<@case'
                                                                                                || $scope.state == '<@continue'
                                                                                                || $scope.state == '<@default'
                                                                                                || $scope.state == '<@else'))
                {
                    // Consume the `/>` characters, set the state, append the characters to the state of the current context, and pop it from the scope chain
                    $position += 2;
                    $state     = 'FusionStartTagSelfClose';
                    $append    = $current + $peek;
                    $pop       = true;
                }
                else
                {
                    // Consume the `/` character
                    $position++;

                    // REGULAR EXPRESSION LITERALS (11.8.5)
                    if ($_lexer_regexp($token, $expression))
                    {
                        // Set the state
                        $state = 'JavaScriptRegExp';

                        // Create the class flag
                        var $class = false;

                        do
                        {
                            // Get the current character from the source
                            $current = $source[$position];

                            // If the current character is an escape character
                            if ($current == '\\')
                            {
                                // Consume the escape character and if there isn't another character or the current character is a line terminator, break
                                if (++$position >= $length || $_newline($source[$position]))
                                    break;
                            }
                            // If the current character is the closing `/` character (and it's not inside a character class)
                            else if (!$class && $current == '/')
                            {
                                // Consume the closing `/`
                                $position++;

                                break;
                            }
                            // If the current character opens a character class (and it's not inside a character class), set the class flag
                            else if (!$class && $current == '[')
                                $class = true;
                            // If the current character closes a character class (and it's inside a character class), reset the class flag
                            else if ($class && $current == ']')
                                $class = false;
                            // LINE TERMINATORS (11.3)
                            else if ($_newline($current))
                                break;
                        }
                        // Continue if the incremented position doesn't exceed the length
                        while (++$position < $length);

                        // If the closing character was a `/`
                        if ($current == '/')
                        {
                            // Get the current character from the source
                            $current = $source[$position];

                            // If the current character starts an identifier
                            if ($_letter($current) || $current == '_' || $current == '$')
                            {
                                // Consume the current character
                                $position++;

                                do
                                {
                                    // Get the current character from the source
                                    $current = $source[$position];

                                    // If the current character isn't a valid identifier character
                                    if (!$_alphanumeric($current) && $current != '_'
                                                                  && $current != '$'
                                                                  && $current != '\u200C'
                                                                  && $current != '\u200D')
                                        // Don't consume the current character
                                        break;
                                }
                                // Continue if the incremented position doesn't exceed the length
                                while (++$position < $length);
                            }
                        }
                    }
                    else
                    {
                        // Set the state
                        $state = 'JavaScriptPunctuator';

                        // If the punctuator is the division assignment operator, consume the `=` character
                        if ($source[$position] == '=')
                            $position++;
                    }
                }
            }
            // NUMERIC LITERALS (11.8.3)
            else if ($_number($current) || $current == '.' && $_number($source[$position + 1]))
            {
                // Set the state
                $state = 'JavaScriptNumber';

                // Get the next character from the source
                var $peek = $current == '0' ?
                            $source[$position + 1] :
                            null;

                // If the first character is non-zero or the zero is followed by a decimal point or exponential indicator
                if ($current != '0' || $peek == '.'
                                    || $peek == 'e'
                                    || $peek == 'E')
                {
                    // If the current character is not a decimal point
                    if ($current != '.')
                    {
                        // Consume the current character
                        $position++;

                        // Consume any other numeric characters
                        while ($_number($source[$position]))
                            $position++;

                        // Get the current character from the source
                        $current = $source[$position];
                    }

                    // If the current character is a decimal point
                    if ($current == '.')
                    {
                        // Consume the `.` character
                        $position++;

                        // Consume any other numeric characters
                        while ($_number($source[$position]))
                            $position++;

                        // Get the current character from the source
                        $current = $source[$position];
                    }

                    // Get the next character from the source
                    $peek = $source[$position + 1];

                    // If the current character is an exponential indicator and the next character is a numeric character (or it is preceded by a valid punctuator)
                    if (($current == 'e' || $current == 'E') && ($_number($peek) || ($peek == '+' || $peek == '-') && $_number($source[$position + 2])))
                    {
                        // Consume the current characters
                        $position += $peek == '+' || $peek == '-' ? 3 : 2;

                        // Consume any other numeric characters
                        while ($_number($source[$position]))
                            $position++;
                    }
                }
                else
                {
                    // Consume the `0` character
                    $position++;

                    // Get the helper function
                    var $helper = $peek == 'b' || $peek == 'B' ?
                                  $_binary :
                                  $peek == 'o' || $peek == 'O' ?
                                  $_octal :
                                  $peek == 'x' || $peek == 'X' ?
                                  $_hexadecimal :
                                  null;

                    // If the next character is in the range of the helper function
                    if ($helper && $helper($source[$position + 1]))
                    {
                        // Consume the flag character and the next character
                        $position += 2;

                        // Consume any other characters in the range of the helper function
                        while ($helper($source[$position]))
                            $position++;
                    }
                }
            }
            // PUNCTUATORS (11.7)
            else if ($current == '{')
            {
                // Consume the current character and set the state
                $position++;
                $state = 'JavaScriptPunctuator';

                // If the lexer has a scope chain
                if ($chain)
                {
                    // Get the identifier if the previous token is a reserved word or the punctuator if it's a punctuator
                    var $identifier = $token && $token.type == 'JavaScriptReservedWord' ?
                                      $token.text() :
                                      null,
                        $punctuator = $token && $token.type == 'JavaScriptPunctuator' ?
                                      $token.text() :
                                      null;

                    // If the previous token is neither an identifier nor a punctuator that preceeds a block
                    if ($identifier && $identifier != 'do'
                                    && $identifier != 'else'
                                    && $identifier != 'export'
                                    && $identifier != 'finally'
                                    && $identifier != 'import'
                                    && $identifier != 'try'
                     || $punctuator && $punctuator != '{'
                                    && $punctuator != ';'
                                    && $punctuator != '=>'
                                    && $punctuator != '}')
                    {
                        // If the previous token is a punctuator that preceeds an object literal
                        if ($punctuator == '('
                         || $punctuator == '['
                         || $punctuator == ','
                         || $punctuator == '?'
                         || $punctuator == ':'
                         || $punctuator == '=')
                            // Push an object key context into the scope chain
                            $push = new Scope('{');
                        // If there is no previous expression or it doesn't preceed a block
                        else if (!$expression || !$_endsWith($expression, '()'))
                        {
                            // Get the next token and its type
                            var $peek = this.peek($position),
                                $type = $peek ?
                                        $peek.type :
                                        null;

                            // If the next token is either an identifier, reserved word, or string literal
                            if ($type == 'JavaScriptIdentifier'
                             || $type == 'JavaScriptReservedWord'
                             || $type == 'JavaScriptDoubleQuotedString'
                             || $type == 'JavaScriptSingleQuotedString')
                            {
                                // Get the token text if it's an identifier
                                var $text = $type == 'JavaScriptIdentifier' ?
                                            $peek.text() :
                                            null;

                                // Get the next token and its type
                                $peek = this.peek($peek.end);
                                $type = $peek ?
                                        $peek.type :
                                        null;

                                // If the next token is the `:` punctuator, push an object key context into the scope chain
                                if ($type == 'JavaScriptPunctuator' && $peek.text() == ':')
                                    $push = new Scope('{');
                                // If the preceeding identifier is either `get` or `set` and the next token is either an identifier or reserved word
                                else if (($text == 'get' || $text == 'set') && ($type == 'JavaScriptIdentifier' || $type == 'JavaScriptReservedWord'))
                                {
                                    // Get the next token and its type
                                    $peek = this.peek($peek.end);
                                    $type = $peek ?
                                            $peek.type :
                                            null;

                                    // If the next token is the `(` punctuator, push an object key context into the scope chain
                                    if ($type == 'JavaScriptPunctuator' && $source[$peek.start] == '(')
                                        $push = new Scope('{');
                                }
                            }
                        }
                    }

                    // If there's a current context and an object key context isn't being pushed into the scope chain, count the current character
                    if ($scope && !$push)
                        $_lexer_count($scope, $current);
                }
            }
            // PUNCTUATORS (11.7)
            else if ($current == '}')
            {
                // Consume the current character and set the state
                $position++;
                $state = 'JavaScriptPunctuator';

                // If there's a current context
                if ($scope)
                {
                    // If the closing brace is unmatched
                    if ($scope.braces == 0)
                    {
                        // If the current character closes a fusion substitution context
                        if ($fusion && $_endsWith($scope.state, '${'))
                        {
                            // Set the state, append the current character to the state of the current context, and pop it from the scope chain
                            $state  = $scope.state == '<${' ?
                                      'FusionSubstitutionClose' :
                                      $scope.state == '@{:${' ?
                                      'FusionObjectSubstitutionClose' :
                                      $scope.state == '@(${' || $scope.state == '@[${' ?
                                      'FusionSelectorSubstitutionClose' :
                                      'FusionStyleSubstitutionClose';
                            $append = $current;
                            $pop    = true;
                        }
                        // If the current character closes either an object key or value context
                        else if ($scope.state == '{' || $scope.state == '{:')
                        {
                            // Replace the state of the current context with an object literal and pop it from the scope chain
                            $replace = '{}';
                            $pop     = true;
                        }
                    }

                    // If the current context isn't being popped from the scope chain, count the current character
                    if (!$pop)
                        $_lexer_count($scope, $current);
                }
            }
            // PUNCTUATORS (11.7)
            else if ($current == '(')
            {
                // Consume the current character and set the state
                $position++;
                $state = 'JavaScriptPunctuator';

                // If the current context is a fusion context
                if ($fusion && $scope && ($_startsWith($scope.state, '<@')
                                       || $_startsWith($scope.state, '</@')))
                {
                    // Get the context of the scope
                    var $context = $scope.state;

                    // If the current character is the opening of a fusion reserved context
                    if ($context == '<@for'
                     || $context == '<@if'
                     || $context == '<@switch'
                     || $context == '<@while'
                     || $context == '<@else for'
                     || $context == '<@else if'
                     || $context == '<@else switch'
                     || $context == '<@else while'
                     || $context == '</@while')
                        // Append the current character to the state of the current context
                        $append = $current;
                }
                // If the lexer has a scope chain and the previous token is a reserved word
                else if ($chain && $token && $token.type == 'JavaScriptReservedWord')
                {
                    // Get the previous token text
                    var $text = $token.text();

                    // If the reserved word must be followed by parentheses
                    if ($text == 'catch'
                     || $text == 'function'
                     || $text == 'for'
                     || $text == 'if'
                   //|| $text == 'let'
                     || $text == 'switch'
                     || $text == 'while'
                     || $text == 'with')
                        // Push a reserved context into the scope chain
                        $push = new Scope($text + '(');
                }

                // If there's a current context and a reserved context isn't being pushed into the scope chain, count the current character
                if ($scope && !$push)
                    $_lexer_count($scope, $current);
            }
            // PUNCTUATORS (11.7)
            else if ($current == ')')
            {
                // Consume the current character and set the state
                $position++;
                $state = 'JavaScriptPunctuator';

                // If the current context is a reserved context
                if ($scope && $_endsWith($scope.state, '('))
                {
                    // If the current context is a fusion context
                    if ($fusion && ($_startsWith($scope.state, '<@')
                                 || $_startsWith($scope.state, '</@')))
                    {
                        // If the current character is the closing of a fusion reserved context, append it to the state of the current context
                        if ($scope.parentheses == 1)
                            $append = $current;
                    }
                    // If the current character closes the reserved context
                    else if ($scope.parentheses == 0)
                    {
                        // Append the current character to the state of the current context, and pop it from the scope chain
                        $append = $current;
                        $pop    = true;
                    }
                }

                // If there's a current context and it isn't being popped from the scope chain, count the current character
                if ($scope && !$pop)
                    $_lexer_count($scope, $current);
            }
            // PUNCTUATORS (11.7)
            else if ($current == '[' || $current == ']')
            {
                // Consume the current character and set the state
                $position++;
                $state = 'JavaScriptPunctuator';

                // If there's a current context in the scope chain, count the current character
                if ($scope)
                    $_lexer_count($scope, $current);
            }
            // PUNCTUATORS (11.7)
            else if ($current == ':')
            {
                // Consume the current character and set the state
                $position++;
                $state = 'JavaScriptPunctuator';

                // If the current character opens an object value context, replace the state of the current context with it
                if ($scope && $scope.braces == 0 && $scope.state == '{')
                    $replace = '{:';
            }
            // PUNCTUATORS (11.7)
            else if ($current == ',')
            {
                // Consume the current character and set the state
                $position++;
                $state = 'JavaScriptPunctuator';

                // If the current character opens an object key context, replace the state of the current context with it
                if ($scope && $scope.braces == 0 && $scope.state == '{:')
                    $replace = '{';
            }
            // PUNCTUATORS (11.7)
            else if ($current == ';'
                  || $current == '?'
                  || $current == '~')
            {
                // Consume the current character and set the state
                $position++;
                $state = 'JavaScriptPunctuator';
            }
            // PUNCTUATORS (11.7)
            else if ($current == '=')
            {
                // Consume the `=` character, get the current character from the source, and set the state
                $current = $source[++$position];
                $state   = 'JavaScriptPunctuator';

                // If the punctuator is the equality operator
                if ($current == '=')
                {
                    // Consume the second `=` character and if the punctuator is the strict equality operator, consume the third `=` character
                    if ($source[++$position] == '=')
                        $position++;
                }
                // If the punctuator is an arrow function, consume the `>` character
                else if ($current == '>')
                    $position++;
            }
            // PUNCTUATORS (11.7)
            else if ($current == '&'
                  || $current == '|'
                  || $current == '+'
                  || $current == '-')
            {
                // Get the next character from the source
                var $peek = $source[$position + 1];

                // Consume the current character (and the next character if the punctuator is an assignment operator, increment/decrement operator, or a logical AND/OR)
                $position += $current == $peek || $peek == '=' ? 2 : 1;
                $state     = 'JavaScriptPunctuator';
            }
            // PUNCTUATORS (11.7)
            else if ($current == '!'
                  || $current == '*'
                  || $current == '^'
                  || $current == '%')
            {
                // Set the state
                $state = 'JavaScriptPunctuator';

                // If the punctuator is an assignment operator
                if ($source[$position + 1] == '=')
                {
                    // Consume both characters
                    $position += 2;

                    // If the punctuator is the strictly negated equality operator, consume the third `=` character
                    if ($current == '!' && $source[$position] == '=')
                        $position++;
                }
                // Consume the current character
                else
                    $position++;
            }
            // PUNCTUATORS (11.7)
            else if ($current == '<' || $current == '>')
            {
                // Get the next character from the source
                var $peek = $source[$position + 1];

                // If the punctuator is either a bitwise shift or a greater than or less than operator
                if ($current == $peek || $peek == '=')
                {
                    // Consume both characters and set the state
                    $position += 2;
                    $state     = 'JavaScriptPunctuator';

                    // If the punctuator is a bitwise shift
                    if ($current == $peek)
                    {
                        // Get the current character from the source
                        $current = $source[$position];

                        // If the punctuator is the zero-fill right shift, consume the current character and get the next character from the source
                        if ($peek == '>' && $current == '>')
                            $current = $source[++$position];

                        // If the punctuator is a bitwise assignment operator, consume the `=` character
                        if ($current == '=')
                            $position++;
                    }
                }
                else
                {
                    // Consume the current character and set the state
                    $position++;
                    $state = 'JavaScriptPunctuator';

                    // If fusion language components are supported
                    if ($fusion)
                    {
                        // If the current context is a fusion context
                        if ($current == '>' && $scope && ($_startsWith($scope.state, '<@')
                                                       || $_startsWith($scope.state, '</@')))
                        {
                            // If the current character is the self closing of a fusion start tag
                            if ($_endsWith($scope.state, '()') || $scope.state == '<@do'
                                                               || $scope.state == '<@else do')
                            {
                                // Set the state, append the current character to the state of the current context, and pop it from the scope chain
                                $state  = $_startsWith($scope.state, '</@') ?
                                          'FusionEndTagClose' :
                                          'FusionStartTagClose';
                                $append = $current;
                                $pop    = true;
                            }
                        }
                        // If the punctuator is a fusion end tag
                        else if ($current == '<' && $peek == '/' && $source[$position + 1] == '@' && $scope && $scope.state == '<@')
                        {
                            // Consume the `/@` characters and set the state
                            $position += 2;
                            $state     = 'FusionEndTagOpen';
                        }
                        // If the current character is the closing of a fusion start tag, set the state
                        else if ($current == '>' && $token && $token.type == 'FusionStartTagOpen')
                            $state = 'FusionStartTagClose';
                        // If the current character is the closing of a fusion end tag, set the state
                        else if ($current == '>' && $token && $token.type == 'FusionEndTagOpen')
                            $state = 'FusionEndTagClose';
                        // If the current characters and context declare an element
                        else if ($current == '<' && $_letter($peek) && $_lexer_regexp($token, $expression))
                        {
                            // Set the state and push the element context into the scope chain
                            $state = 'HTMLStartTagOpen';
                            $push  = new Scope('<');

                            // Count the opening tag in the element context
                            $_lexer_count($push, $current);
                        }
                    }
                    // If the punctuator opens a script end tag
                    else if ($current == '<' && $peek == '/' && $source.substr($position + 1, 6) == 'script' && $scope && $scope.state == '<script')
                    {
                        // Consume the `/` character, set the state, append the `>` character to the state of the current context and pop it from the scope chain
                        $position++;
                        $state  = 'HTMLEndTagOpen';
                        $append = '>';
                        $pop    = true;
                    }
                }
            }
            // PUNCTUATORS (11.7)
            else if ($current == '.')
            {
                // Consume the `.` character, get the current character from the source, and set the state
                $current = $source[++$position];
                $state   = 'JavaScriptPunctuator';

                // If the punctuator is the spread operator, consume both `.` characters
                if ($current == '.' && $source[$position + 1] == '.')
                    $position += 2;
            }
            // If the current character opens a fusion context
            else if ($fusion && $current == '@')
            {
                // Get the next character from the source
                var $peek = $source[$position + 1];

                // Consume the current character and set the state
                $position++;
                $state = $peek == '{' ?
                         'FusionObject' :
                         $peek == '(' || $peek == '[' ?
                         'FusionSelector' :
                         'JavaScriptInvalidCharacter';

                // If the current character is either a fusion style or selector context
                if ($chain && ($peek == '{'
                            || $peek == '('
                            || $peek == '['))
                    // Push either a fusion style or selector context into the scope chain
                    $push = new Scope($current + $peek);
            }
            else
            {
                // Consume the current character and set the state
                $position++;
                $state = 'JavaScriptInvalidCharacter';
            }

            break;

        // ---------- HTML ----------
        case 'html':
        case 'fhtml':
        case 'HTMLStartTagClose':
        case 'HTMLStartTagSelfClose':
        case 'HTMLEndTagClose':
        case 'HTMLCharacterReference':
        case 'HTMLCommentClose':
        case 'HTMLBogusCommentClose':
        case 'HTMLDOCTYPEClose':
        case 'HTMLCDATAClose':
        case 'FusionStartTagClose':
        case 'FusionStartTagSelfClose':
        case 'FusionEndTagClose':
        case 'FusionDirectiveTagClose':
        case 'FusionSubstitutionClose':

            // DATA STATE (12.2.4.1)
            $state = 'HTMLText';

        case 'HTMLText':

            do
            {
                // Get the current character from the source
                var $current = $source[$position];

                // TAG OPEN STATE (12.2.4.8)
                if ($current == '<')
                {
                    // Get the next character from the source
                    var $peek = $source[$position + 1];

                    // TAG NAME STATE (12.2.4.10)
                    if ($_letter($peek))
                    {
                        // If these aren't the first characters, break (and don't consume them)
                        if ($position != $start)
                            break;

                        // Consume the `<` character and set the state
                        $position++;
                        $state = 'HTMLStartTagOpen';

                        // If there's a current context in the scope chain, count the current character
                        if ($scope)
                            $_lexer_count($scope, $current);

                        break;
                    }
                    // END TAG OPEN STATE (12.2.4.9)
                    else if ($peek == '/')
                    {
                        // If `</` aren't the first characters, break (and don't consume them)
                        if ($position != $start)
                            break;

                        // Get the next character from the source
                        $peek = $source[$position + 2];

                        // If the closing tag is a fusion end tag
                        if ($fusion && $peek == '@')
                        {
                            // Consume the `</@` characters and set the state
                            $position += 3;
                            $state     = 'FusionEndTagOpen';
                        }
                        // TAG NAME STATE (12.2.4.10)
                        else if ($_letter($peek))
                        {
                            // Consume the `</` characters and set the state
                            $position += 2;
                            $state     = 'HTMLEndTagOpen';
                        }
                        // BOGUS COMMENT STATE (12.2.4.44)
                        else
                        {
                            // Consume the `</` characters and set the state
                            $position += 2;
                            $state     = 'HTMLBogusCommentOpen';
                        }

                        break;
                    }
                    // If the opening tag is a fusion start tag
                    else if ($fusion && $peek == '@')
                    {
                        // If `<@` aren't the first characters, break (and don't consume them)
                        if ($position != $start)
                            break;

                        // Consume the `<@` characters and set the state
                        $position += 2;
                        $state     = 'FusionStartTagOpen';

                        break;
                    }
                    // MARKUP DECLARATION OPEN STATE (12.2.4.45)
                    else if ($peek == '!')
                    {
                        // If `<!` aren't the first characters, break (and don't consume them)
                        if ($position != $start)
                            break;

                        // Get the next character from the source
                        $peek = $source[$position + 2];

                        // DATA STATE (12.2.4.1)
                        if ($peek == '>')
                        {
                            // Consume the `<!` characters and set the state
                            $position += 2;
                            $state     = 'HTMLBogusCommentOpen';
                        }
                        // If the opening tag is a fusion directive tag
                        else if ($fusion && $peek == '@')
                        {
                            // Consume the `<!@` characters and set the state
                            $position += 3;
                            $state     = 'FusionDirectiveTagOpen';
                        }
                        // COMMENT START STATE (12.2.4.46)
                        else if ($peek == '-' && $source[$position + 3] == '-')
                        {
                            // Consume the `<!--` characters and set the state
                            $position += 4;
                            $state     = 'HTMLCommentOpen';
                        }
                        else
                        {
                            // Get the next string of seven characters
                            $peek = $source.substr($position + 2, 7);

                            // DOCTYPE STATE (12.2.4.52)
                            if ($peek == 'DOCTYPE')
                            {
                                // Consume the `<!DOCTYPE` characters and set the state
                                $position += 9;
                                $state     = 'HTMLDOCTYPEOpen';
                            }
                            // CDATA SECTION STATE (12.2.4.68)
                            else if ($peek == '[CDATA[')
                            {
                                // Consume the `<![CDATA[` characters and set the state
                                $position += 9;
                                $state     = 'HTMLCDATAOpen';
                            }
                            // BOGUS COMMENT STATE (12.2.4.44)
                            else
                            {
                                // Consume the `<!` characters and set the state
                                $position += 2;
                                $state     = 'HTMLBogusCommentOpen';
                            }
                        }

                        break;
                    }
                    // BOGUS COMMENT STATE (12.2.4.44)
                    else if ($peek == '?')
                    {
                        // If `<?` aren't the first characters, break (and don't consume them)
                        if ($position != $start)
                            break;

                        // Consume the `<?` characters and set the state
                        $position += 2;
                        $state     = 'HTMLBogusCommentOpen';

                        break;
                    }
                }
                // CHARACTER REFERENCE IN DATA STATE (12.2.4.2)
                else if ($current == '&')
                {
                    // Declare the helper function and get the next character from the source
                    var $helper = null,
                        $peek   = $source[$position + 1];

                    // If the character reference is a unicode character code
                    if ($peek == '#')
                    {
                        // Get the next character from the source
                        $peek = $source[$position + 2];

                        // If the character code is hexadecimal
                        if ($peek == 'x' || $peek == 'X')
                        {
                            // Get the next character from the source
                            $peek = $source[$position + 3];

                            // If the character code is hexadecimal
                            if ($_hexadecimal($peek))
                            {
                                // If these aren't the first characters, break (and don't consume them)
                                if ($position != $start)
                                    break;

                                // Consume the current characters and set the helper function
                                $position += 4;
                                $helper    = $_hexadecimal;
                            }
                        }
                        // If the character code is numeric
                        else if ($_number($peek))
                        {
                            // If these aren't the first characters, break (and don't consume them)
                            if ($position != $start)
                                break;

                            // Consume the current characters and set the helper function
                            $position += 3;
                            $helper    = $_number;
                        }
                    }
                    // If the character matches a named character reference
                    else if ($_letter($peek))
                    {
                        // If these aren't the first characters, break (and don't consume them)
                        if ($position != $start)
                            break;

                        // Consume the current characters and set the helper function
                        $position += 2;
                        $helper    = $_alphanumeric;
                    }

                    // If this isn't a character reference, continue (and consume the `&` character)
                    if (!$helper)
                        continue;

                    // Set the state
                    $state = 'HTMLCharacterReference';

                    // Consume any other characters in the range of the helper function
                    while ($helper($source[$position]))
                        $position++;

                    // If the next consumed character isn't a semi-colon, unconsume it
                    if ($source[$position++] != ';')
                        $position--;

                    break;
                }
                // If the current characters open a fusion substitution
                else if ($fusion && $current == '$' && $source[$position + 1] == '{' && $scope && $scope.state == '<')
                {
                    // If these aren't the first characters, break (and don't consume them)
                    if ($position != $start)
                        break;

                    // Consume the `${` characters, set the state, and push the substitution context into the scope chain
                    $position += 2;
                    $state     = 'FusionSubstitutionOpen';
                    $push      = new Scope('<${');

                    break;
                }
            }
            // Continue if the incremented position doesn't exceed the length
            while (++$position < $length);

            break;

        case 'HTMLStartTagOpen':
        case 'HTMLEndTagOpen':

            // TAG NAME STATE (12.2.4.10)
            if ($_letter($source[$position]))
            {
                // Consume the current character and set the state
                $position++;
                $state = $state != 'HTMLStartTagOpen' ?
                         'HTMLEndTagName' :
                         'HTMLStartTagName';

                do
                {
                    // Get the current character from the source
                    var $current = $source[$position];

                    // If the current character isn't allowed in a tag name, break (and don't consume it)
                    if ($current == '>' || $current == '/' || $_space($current))
                        break;
                }
                // Continue if the incremented position doesn't exceed the length
                while (++$position < $length);

                // If there's a current context in the scope chain, set the current tag name
                if ($scope)
                    $scope.tag = $source.substring($start, $position);

                break;
            }

        case 'HTMLEndTagName':
        case 'HTMLEndTagText':
        case 'HTMLEndTagWhitespace':

            // Set the end flag
            $end = $state != 'HTMLStartTagOpen';

        case 'HTMLStartTagName':
        case 'HTMLStartTagSolidus':
        case 'HTMLStartTagWhitespace':
        case 'HTMLAttributeName':
        case 'HTMLAttributeOperator':
        case 'HTMLAttributeValue':
        case 'HTMLAttributeDoubleQuotedValue':
        case 'HTMLAttributeSingleQuotedValue':
        case 'FusionAttributeTemplateString':
        case 'FusionAttributeTemplateStringTail':

            // Get the current character from the source
            var $current = $source[$position];

            // DATA STATE (12.2.4.1)
            if ($current == '>')
            {
                // Consume the `>` character and set the state
                $position++;
                $state = $end ?
                         'HTMLEndTagClose' :
                         'HTMLStartTagClose';

                // If there's a current context in the scope chain
                if ($scope)
                {
                    // If the tag is either an end tag or void tag, count the current character
                    if ($end || $_lexer_void($scope.tag))
                        $_lexer_count($scope, $current);

                    // If the tag being closed is either a script or style tag, push either a script or style context into the scope chain
                    if (!$end && (!$fusion && $scope.tag == 'script' || $scope.tag == 'style'))
                        $push = new Scope('<' + $scope.tag);

                    // Reset the current tag name
                    $scope.tag = '';
                }
            }
            // If the current character is whitespace
            else if ($_space($current))
            {
                // Consume the whitespace character and set the state
                $position++;
                $state = $end ?
                         'HTMLEndTagWhitespace' :
                         'HTMLStartTagWhitespace';

                // Consume any other whitespace characters
                while ($_space($source[$position]))
                    $position++;
            }
            // If the tag is an end tag
            else if ($end)
            {
                // Consume the current character and set the state
                $position++;
                $state = 'HTMLEndTagText';

                do
                {
                    // Get the current character from the source
                    $current = $source[$position];

                    // If the current character isn't allowed in an end tag text, break (and don't consume it)
                    if ($current == '>' || $_space($current))
                        break;
                }
                // Continue if the incremented position doesn't exceed the length
                while (++$position < $length);
            }
            // BEFORE ATTRIBUTE VALUE STATE (12.2.4.37)
            else if ($current == '=' && $token && $token.type == 'HTMLAttributeName')
            {
                // Consume the `=` character and set the state
                $position++;
                $state = 'HTMLAttributeOperator';
            }
            // ATTRIBUTE VALUE (DOUBLE-QUOTED/SINGLE-QUOTED/UNQUOTED) STATE (12.2.4.38/12.2.4.39/12.2.4.40)
            else if ($token && $token.type == 'HTMLAttributeOperator')
            {
                // Get the break character
                var $break = $current == '"'
                          || $current == "'"
                          || $current == '`' && $fusion && $scope && $scope.state == '<' ?
                             $current :
                             null;

                // Consume the current character and set the state
                $position++;
                $state = $break == '`' ?
                         'FusionAttributeTemplateString' :
                         $break == '"' ?
                         'HTMLAttributeDoubleQuotedValue' :
                         $break == "'" ?
                         'HTMLAttributeSingleQuotedValue' :
                         'HTMLAttributeValue';

                do
                {
                    // Get the current character from the source
                    $current = $source[$position];

                    // If a break character is set
                    if ($break)
                    {
                        // If the current character is the break character
                        if ($current == $break)
                        {
                            // Consume the break character
                            $position++;

                            break;
                        }
                        // If the state is a template string
                        else if ($break == '`')
                        {
                            // If the current character opens a template substitution
                            if ($current == '$' && $source[$position + 1] == '{')
                            {
                                // Consume the `${` characters, set the state, and push a substitution context into the scope chain
                                $position += 2;
                                $state     = 'FusionAttributeTemplateStringHead';
                                $push      = new Scope('=`${');

                                break;
                            }
                        }
                    }
                    // If the current character isn't allowed in an unquoted attribute value, break (and don't consume it)
                    else if ($current == '>' || $_space($current))
                        break;
                }
                // Continue if the incremented position doesn't exceed the length
                while (++$position < $length);
            }
            // SELF-CLOSING START TAG STATE (12.2.4.43)
            else if ($current == '/')
            {
                // Consume the `/` character, get the current character, and set the state
                $current = $source[++$position];
                $state   = 'HTMLStartTagSolidus';

                // DATA STATE (12.2.4.1)
                if ($current == '>')
                {
                    // Consume the `>` character and set the state
                    $position++;
                    $state = 'HTMLStartTagSelfClose';

                    // If there's a current context in the scope chain
                    if ($scope)
                    {
                        // Count the current character
                        $_lexer_count($scope, $current);

                        // Reset the current tag name
                        $scope.tag = '';
                    }
                }
                // Consume any other solidus characters
                else while ($current == '/')
                    $current = $source[++$position];
            }
            // ATTRIBUTE NAME STATE (12.2.4.35)
            // BEFORE/AFTER ATTRIBUTE NAME STATE (12.2.4.34/12.2.4.36)
            // CHARACTER REFERENCE IN ATTRIBUTE VALUE STATE (12.2.4.41)
            // AFTER ATTRIBUTE VALUE (QUOTED) STATE (12.2.4.42)
            else
            {
                // Consume the current character and set the state
                $position++;
                $state = 'HTMLAttributeName';

                do
                {
                    // Get the current character from the source
                    $current = $source[$position];

                    // If the current character isn't allowed in an attribute name, break (and don't consume it)
                    if ($current == '>' || $current == '/' || $current == '=' || $_space($current))
                        break;
                }
                // Continue if the incremented position doesn't exceed the length
                while (++$position < $length);
            }

            break;

        case 'HTMLCommentOpen':

            // COMMENT STATE (12.2.4.48)
            $state = 'HTMLCommentText';

        case 'HTMLCommentText':

            do
            {
                // Get the current character from the source
                var $current = $source[$position];

                // DATA STATE (12.2.4.1)
                if ($current == '>')
                {
                    // If `>` isn't the first character or the previous token did not open the comment, continue (and consume the `>` character)
                    if ($position != $start || !$token || $token.type != 'HTMLCommentOpen')
                        continue;

                    // Consume the `>` character and set the state
                    $position++;
                    $state = 'HTMLCommentClose';

                    break;
                }
                // COMMENT START/END DASH STATE (12.2.4.47/12.2.4.49)
                else if ($current == '-')
                {
                    // Get the next character from the source
                    var $peek = $source[$position + 1];

                    // DATA STATE (12.2.4.1)
                    if ($peek == '>')
                    {
                        // If `->` aren't the first characters or the previous token did not open the comment, continue (and consume the `-` character)
                        if ($position != $start || !$token || $token.type != 'HTMLCommentOpen')
                            continue;

                        // Consume the `->` characters and set the state
                        $position += 2;
                        $state     = 'HTMLCommentClose';

                        break;
                    }
                    // COMMENT END STATE (12.2.4.50)
                    else if ($peek == '-')
                    {
                        // Get the next character from the source
                        $peek = $source[$position + 2];

                        // DATA STATE (12.2.4.1)
                        // COMMENT END BANG STATE (12.2.4.51)
                        if ($peek == '>' || $peek == '!' && $source[$position + 3] == '>')
                        {
                            // If these aren't the first characters, break (and don't consume them)
                            if ($position != $start)
                                break;

                            // Consume the current characters and set the state
                            $position += $peek == '!' ? 4 : 3;
                            $state     = 'HTMLCommentClose';

                            break;
                        }
                    }
                }
            }
            // Continue if the incremented position doesn't exceed the length
            while (++$position < $length);

            break;

        case 'HTMLBogusCommentOpen':

            // BOGUS COMMENT STATE (12.2.4.44)
            $state = 'HTMLBogusCommentText';

        case 'HTMLBogusCommentText':

            do
            {
                // DATA STATE (12.2.4.1)
                if ($source[$position] == '>')
                {
                    // If `>` isn't the first character, break (and don't consume it)
                    if ($position != $start)
                        break;

                    // Consume the `>` character and set the state
                    $position++;
                    $state = 'HTMLBogusCommentClose';

                    break;
                }
            }
            // Continue if the incremented position doesn't exceed the length
            while (++$position < $length);

            break;

        case 'HTMLDOCTYPEOpen':
        case 'HTMLDOCTYPEString':
        case 'HTMLDOCTYPEDoubleQuotedString':
        case 'HTMLDOCTYPESingleQuotedString':
        case 'HTMLDOCTYPEWhitespace':

            // Get the current character from the source
            var $current = $source[$position];

            // DATA STATE (12.2.4.1)
            if ($current == '>')
            {
                // Consume the `>` character and set the state
                $position++;
                $state = 'HTMLDOCTYPEClose';
            }
            // If the current character is whitespace
            else if ($_space($current))
            {
                // Consume the space character and set the state
                $position++;
                $state = 'HTMLDOCTYPEWhitespace';

                // Consume any other whitespace characters
                while ($_space($source[$position]))
                    $position++;
            }
            else
            {
                // Get the break character
                var $break = $current == '"' || $current == "'" ?
                             $current :
                             null;

                // Consume the current character and set the state
                $position++;
                $state = $current == '"' ?
                         'HTMLDOCTYPEDoubleQuotedString' :
                         $current == "'" ?
                         'HTMLDOCTYPESingleQuotedString' :
                         'HTMLDOCTYPEString';

                do
                {
                    // Get the current character from the source
                    $current = $source[$position];

                    // If a break character is set
                    if ($break)
                    {
                        // If the current character is the break character
                        if ($current == $break)
                        {
                            // Consume the break character
                            $position++;

                            break;
                        }
                    }
                    // If the current character isn't allowed in a DOCTYPE string, break (and don't consume it)
                    else if ($current == '>' || $_space($current))
                        break;
                }
                // Continue if the incremented position doesn't exceed the length
                while (++$position < $length);
            }

            break;

        case 'HTMLCDATAOpen':

            // CDATA SECTION STATE (12.2.4.68)
            $state = 'HTMLCDATAText';

        case 'HTMLCDATAText':

            do
            {
                // DATA STATE (12.2.4.1)
                if ($source[$position] == ']' && $source[$position + 1] == ']' && $source[$position + 2] == '>')
                {
                    // If `]]>` aren't the first characters, break (and don't consume them)
                    if ($position != $start)
                        break;

                    // Consume the `]]>` characters and set the state
                    $position += 3;
                    $state     = 'HTMLCDATAClose';

                    break;
                }
            }
            // Continue if the incremented position doesn't exceed the length
            while (++$position < $length);

            break;

        // ---------- CSS ----------
        case 'css':
        case 'fcss':
        case 'CSSIdentifier':
        case 'CSSNumber':
        case 'CSSDelimiter':
        case 'CSSPunctuator':
        case 'CSSDimension':
        case 'CSSPercentage':
        case 'CSSComma':
        case 'CSSColon':
        case 'CSSSemicolon':
        case 'CSSDoubleQuotedString':
        case 'CSSSingleQuotedString':
        case 'CSSHash':
        case 'CSSFunction':
        case 'CSSAtKeyword':
        case 'CSSMatch':
        case 'CSSUrl':
        case 'CSSColumn':
        case 'CSSUnicodeRange':
        case 'CSSComment':
        case 'CSSWhitespace':
        case 'FusionObject':
        case 'FusionSelector':
        case 'FusionObjectSubstitutionClose':
        case 'FusionSelectorSubstitutionClose':
        case 'FusionStyleSubstitutionClose':

            // Get the current character from the source and check if it's numeric
            var $current = $source[$position],
                $numeric = $_number($current);

            // If the current character isn't a number but could be numeric
            if (!$numeric && ($current == '+'
                           || $current == '-'
                           || $current == '.'))
            {
                // Get the next character from the source
                var $peek = $source[$position + 1];

                // NUMBER (4.3.11)
                $numeric = $_number($peek) || $current != '.' && $peek == '.' && $_number($source[$position + 2]);
            }

            // NUMERIC (4.3.3)
            if ($numeric)
            {
                // Set the state
                $state = 'CSSNumber';

                // If the current character is not a decimal point
                if ($current != '.')
                {
                    // Consume the current character
                    $position++;

                    // Consume any other numeric characters
                    while ($_number($source[$position]))
                        $position++;

                    // Get the current character from the source
                    $current = $source[$position];
                }

                // If the current character is a decimal point
                if ($current == '.' && $_number($source[$position + 1]))
                {
                    // Consume the current characters
                    $position += 2;

                    // Consume any other numeric characters
                    while ($_number($source[$position]))
                        $position++;

                    // Get the current character from the source
                    $current = $source[$position];
                }

                // Get the next character from the source
                var $peek = $source[$position + 1];

                // If the current character is an exponential indicator and the next character is a numeric character (or it is preceded by a valid punctuator)
                if (($current == 'e' || $current == 'E') && ($_number($peek) || ($peek == '+' || $peek == '-') && $_number($source[$position + 2])))
                {
                    // Consume the current characters
                    $position += $peek == '+' || $peek == '-' ? 3 : 2;

                    // Consume any other numeric characters
                    while ($_number($source[$position]))
                        $position++;

                    // Get the current character from the source
                    $current = $source[$position];
                }

                // PERCENTAGE (4.3.3)
                if ($current == '%')
                {
                    // Consume the `%` character and set the state
                    $position++;
                    $state = 'CSSPercentage';

                    break;
                }
                // If the current character can't start an identifier
                else if (!$_letter($current) && $current != '_'
                                             && $current != '\\'
                                             && $current != '-'
                                             && $current <  '\u0080')
                    // Don't consume the current character
                    break;
            }

            // DIMENSION (4.3.3)
            // IDENT (4.3.10)
            if ($numeric || $_letter($current) || $current == '_'
                                               || $current == '\\'
                                               || $current == '#'
                                               || $current == '@'
                                               || $current == '-'
                                               || $current >= '\u0080')
            {
                // UNICODE-RANGE (4.3.7)
                if (!$numeric && ($current == 'u'
                               || $current == 'U') && $source[$position + 1] == '+' && ($_hexadecimal($source[$position + 2]) || $source[$position + 2] == '?'))
                {
                    // Consume the current characters and set the state
                    $position += 2;
                    $state = 'CSSUnicodeRange';

                    // Create the character count
                    var $count = 0;

                    // Count any six hexadecimal characters
                    while ($count < 6 && $_hexadecimal($source[$position + $count]))
                        $count++;

                    // Get the next character from the source
                    var $peek = $source[$position + $count];

                    // If the unicode range is a range
                    if ($peek == '-' && $_hexadecimal($source[$position + $count + 1]))
                    {
                        // Consume the counted hexadecimal characters along with the `-` character and reset the character count
                        $position += $count + 1;
                        $count     = 0;

                        // Count any six hexadecimal characters
                        while ($count < 6 && $_hexadecimal($source[$position + $count]))
                            $count++;
                    }
                    // If the unicode range is a 0-F range, count any six minus the character count `?` characters
                    else if ($peek == '?')
                        while ($count < 6 && $source[$position + $count] == '?')
                            $count++;

                    // Consume the counted characters
                    $position += $count;
                }
                // IDENT-LIKE (4.3.4)
                // HASH (4.3.12)
                // AT-KEYWORD (4.3.12)
                else
                {
                    // Set the state
                    $state = $numeric ?
                             'CSSDimension' :
                             $current == '#' ?
                             'CSSHash' :
                             $current == '@' ?
                             'CSSAtKeyword' :
                             'CSSIdentifier';

                    // If the current character is an escape character
                    if ($current == '\\')
                    {
                        // If the next character is a newline
                        if ($_newline($source[$position + 1]))
                        {
                            // If the identifier isn't numeric, consume the delimiter character
                            if (!$numeric)
                                $position++;

                            // Set the state
                            $state = $numeric ?
                                     'CSSNumber' :
                                     'CSSDelimiter';

                            break;
                        }

                        // Consume the escape character and if there is another character, consume it
                        if (++$position < $length)
                            $position++;
                    }
                    // If the current character is a delimiter character
                    else if ($current == '#'
                          || $current == '@'
                          || $current == '-')
                    {
                        // Get the next character from the source
                        var $peek = $source[$position + 1];

                        // If the next character isn't a valid starting identifier character
                        if (!$_letter($peek) &&  $peek != '_'
                                             &&  $peek != '-'
                                             && ($peek != '\\' || $_newline($source[$position + 2])) && $peek < '\u0080' && ($current != '#' || !$_number($peek)))
                        {
                            // If the identifier isn't numeric, consume the delimiter character
                            if (!$numeric)
                                $position++;

                            // Set the state
                            $state = $numeric ?
                                     'CSSNumber' :
                                     'CSSDelimiter';

                            break;
                        }
                        // If the current character is an at-keyword and the next character could be a delimiter
                        else if ($current == '@' && $peek == '-')
                        {
                            // Get the next character from the source
                            $peek = $source[$position + 2];

                            // If the next character isn't a valid starting identifier character
                            if (!$_letter($peek) &&  $peek != '_'
                                                 &&  $peek != '-'
                                                 && ($peek != '\\' || $_newline($source[$position + 3])) && $peek < '\u0080')
                            {
                                // Consume the delimiter character and set the state
                                $position++;
                                $state = 'CSSDelimiter';

                                break;
                            }

                            // Consume the current characters
                            $position += 3;
                        }
                        // Consume the current characters
                        else
                            $position += 2;
                    }
                    // Consume the current character
                    else
                        $position++;

                    do
                    {
                        // Get the current character from the source
                        $current = $source[$position];

                        // If the current character is an escape character
                        if ($current == '\\')
                        {
                            // If the next character is a newline, break (and don't consume the escape character)
                            if ($_newline($source[$position + 1]))
                                break;

                            // Consume the escape character and if there isn't another character, break
                            if (++$position >= $length)
                                break;
                        }
                        // If the current character isn't a valid identifier character
                        else if (!$_alphanumeric($current) && $current != '_'
                                                           && $current != '-'
                                                           && $current < '\u0080')
                            // Don't consume the current character
                            break;
                    }
                    // Continue if the incremented position doesn't exceed the length
                    while (++$position < $length);

                    // FUNCTION (4.3.4)
                    if ($current == '(' && $state == 'CSSIdentifier')
                    {
                        // Consume the `(` character and set the state
                        $position++;
                        $state = 'CSSFunction';

                        // If there's a current context, count the current character
                        if ($scope)
                            $_lexer_count($scope, $current);

                        // URL (4.3.4)
                        if ($position == $start + 4 && $source.substr($start, 3).toLowerCase() == 'url')
                        {
                            // Create the whitespace count
                            var $count = 0;

                            // Count any other whitespace characters
                            while ($_space($source[$position + $count]))
                                $count++;

                            // Get the next character from the source
                            var $peek = $source[$position + $count];

                            // If the next character isn't a string
                            if ($peek != '"' && $peek != "'")
                            {
                                // Consume the whitespace characters and set the state
                                $position += $count;
                                $state     = 'CSSUrl';

                                do
                                {
                                    // Get the current character from the source
                                    $current = $source[$position];

                                    // If the current character is a closing parentheses
                                    if ($current == ')')
                                    {
                                        // Consume the `)` character
                                        $position++;

                                        break;
                                    }
                                    // If the current character is an escape character, consume it and break if there isn't another character
                                    else if ($current == '\\' && ++$position >= $length)
                                        break;
                                }
                                // Continue if the incremented position doesn't exceed the length
                                while (++$position < $length);
                            }
                        }
                    }
                }
            }
            // COMMENTS (4.3.2)
            else if ($current == '/' && $source[$position + 1] == '*')
            {
                // Consume the `/*` characters and set the state
                $position += 2;
                $state     = 'CSSComment';

                do
                {
                    // If the current characters close the comment
                    if ($source[$position] == '*' && $source[$position + 1] == '/')
                    {
                        // Consume the `*/` characters
                        $position += 2;

                        break;
                    }
                }
                // Continue if the incremented position doesn't exceed the length
                while (++$position < $length);
            }
            // WHITESPACE (4.3.1)
            else if ($_space($current))
            {
                // Consume the whitespace character and set the state
                $position++;
                $state = 'CSSWhitespace';

                // Consume any other whitespace characters
                while ($_space($source[$position]))
                    $position++;
            }
            // STRING (4.3.5)
            else if ($current == '"' || $current == "'")
            {
                // Get the break character
                var $break = $current;

                // Consume the current character and set the state
                $position++;
                $state = $break == '"' ?
                         'CSSDoubleQuotedString' :
                         'CSSSingleQuotedString';

                do
                {
                    // Get the current character from the source
                    $current = $source[$position];

                    // If the current character is an escape sequence
                    if ($current == '\\')
                    {
                        // Consume the escape character and if there isn't another character, break
                        if (++$position >= $length)
                            break;

                        // If the current character is a carriage return and the next character is a line break, consume the carriage return
                        if ($source[$position] == '\r' && $source[$position + 1] == '\n')
                            $position++;
                    }
                    // If the current character is the break character
                    else if ($current == $break)
                    {
                        // Consume the current character
                        $position++;

                        break;
                    }
                    // LINE TERMINATORS (11.3)
                    else if ($_newline($current))
                        break;
                }
                // Continue if the incremented position doesn't exceed the length
                while (++$position < $length);
            }
            // SUFFIX/SUBSTRING/PREFIX/DASH/INCLUDE MATCH (4.3.1)
            // COLUMN (4.3.1)
            else if ($current == '$'
                  || $current == '*'
                  || $current == '^'
                  || $current == '|'
                  || $current == '~')
            {
                // Get the next character from the source
                var $peek = $source[$position + 1];

                // If the current characters open a declaration substitution
                if ($fusion && $current == '$' && $peek == '{' && $scope && ($scope.state == '@{:'
                                                                          || $scope.state == '@('
                                                                          || $scope.state == '@['
                                                                          || $scope.state == '<style'))
                {
                    // Consume the current characters, set the state, and push a declaration substitution context into the scope chain
                    $position += 2;
                    $state     = $scope.state == '@{:' ?
                                 'FusionObjectSubstitutionOpen' :
                                 $scope.state == '@(' || $scope.state == '@[' ?
                                 'FusionSelectorSubstitutionOpen' :
                                 'FusionStyleSubstitutionOpen';
                    $push      = new Scope($scope.state + $current + $peek);
                }
                // If the current character is a column
                else if ($current == '|' && $peek == '|')
                {
                    // Consume the current characters and set the state
                    $position += 2;
                    $state     = 'CSSColumn';
                }
                // If the current characters are a match
                else if ($peek == '=')
                {
                    // Consume the current characters and set the state
                    $position += 2;
                    $state     = 'CSSMatch';
                }
                else
                {
                    // Consume the current character and set the state
                    $position++;
                    $state = 'CSSDelimiter';
                }
            }
            // PARENTHESIS/BRACKET/CURLY BRACKET (4.3.1)
            else if ($current == '('
                  || $current == ')'
                  || $current == '['
                  || $current == ']'
                  || $current == '{'
                  || $current == '}')
            {
                // Consume the current character and set the state
                $position++;
                $state = 'CSSPunctuator';

                // If there's a current context in the scope chain, count the current character
                if ($scope)
                    $_lexer_count($scope, $current);
            }
            // COMMA (4.3.1)
            else if ($current == ',')
            {
                // Consume the current character and set the state
                $position++;
                $state = 'CSSComma';
            }
            // COLON (4.3.1)
            else if ($current == ':')
            {
                // Consume the current character and set the state
                $position++;
                $state = 'CSSColon';

                // If the current context is a declaration name context, append the current character to make it a declaration value context
                if ($scope && $scope.state == '@{')
                    $append = $current;
            }
            // SEMICOLON (4.3.1)
            else if ($current == ';')
            {
                // Consume the current character and set the state
                $position++;
                $state = 'CSSSemicolon';

                // If the current context is a declaration value context, replace it with a declaration name context
                if ($scope && $scope.state == '@{:')
                    $replace = '@{';
            }
            // COMMENT (4.3.1)
            else if ($current == '<')
            {
                // Consume the `<` character, get the current character, and set the state
                $current = $source[++$position];
                $state   = 'CSSDelimiter';

                // If the current character opens a markup comment
                if ($current == '!' && $source[$position + 1] == '-' && $source[$position + 2] == '-' && $scope && $scope.state == '<style')
                {
                    // Consume the `!--` characters and set the state
                    $position += 3;
                    $state     = 'CSSComment';

                    do
                    {
                        // If the current character closes the comment
                        if ($source[$position] == '-' && $source[$position + 1] == '-' && $source[$position + 2] == '>')
                        {
                            // Consume the `-->` characters
                            $position += 3;

                            break;
                        }
                    }
                    // Continue if the incremented position doesn't exceed the length
                    while (++$position < $length);
                }
                // If the punctuator is a markup end tag
                else if ($current == '/' && $source.substr($position + 1, 5) == 'style' && $scope && $scope.state == '<style')
                {
                    // Consume the `/` character, set the state, append the `>` character to the state of the current context and pop it from the scope chain
                    $position++;
                    $state  = 'HTMLEndTagOpen';
                    $append = '>';
                    $pop    = true;
                }
            }
            else
            {
                // Consume the current character and set the state
                $position++;
                $state = 'CSSDelimiter';
            }

            break;

        default:

            return null;
    }

    // Set the current position and state
    this.position = $position;
    this.state    = $state;

    // If the element state is being closed
    if ($fusion && $scope && $scope.state == '<' && $scope.tags == 0)
    {
        // Append the `>` character to the state of the current context and pop it from the scope chain
        $append = '>';
        $pop    = true;

        // Set the language state
        this.state = $language;
    }
    // If either the style or selector states are being closed
    else if ($fusion && $scope && (($scope.state == '@{'
                                 || $scope.state == '@{:') && $scope.braces      == 0
                                 || $scope.state == '@('   && $scope.parentheses == 0
                                 || $scope.state == '@['   && $scope.brackets    == 0))
    {
        // Replace the state of the current context and pop it from the scope chain
        $pop     = true;
        $replace = $scope.state == '@(' ?
                   '@()' :
                   $scope.state == '@[' ?
                   '@[]' :
                   '@{}';

        // Set the language state
        this.state = $language;
    }
    // If a closing start tag is pushing a context into the scope chain
    else if ($push && $state == 'HTMLStartTagClose')
    {
        // If a script context is being pushed, set the language state
        if ($push.state == '<script')
            this.state = 'js';
        // If a style context is being pushed, set the language state
        else if ($push.state == '<style')
            this.state = 'css';
    }

    // If there's a current context
    if ($scope)
    {
        // If there's a replacement string for the state, set the state of the current context
        if ($replace)
            $scope.state = $replace;

        // If there's an append string for the state, append the string to the state of the current context
        if ($append)
            $scope.state += $append;
    }

    // If there's a scope chain
    if ($chain)
    {
        // If a context should be pushed into the scope chain
        if ($push)
        {
            // If the pop flag is set and there are already contexts in the scope chain, replace the last context
            if ($pop && $scopes)
                $chain[$scopes - 1] = $push;
            // Push the context into the scope chain
            else
                $chain.push($push);
        }
        // If the pop flag is set, pop the current context from the scope chain
        else if ($pop)
            $chain.pop();
    }

    // If no characters were consumed, return null
    if ($position == $start)
        return null;

    // Create the current token
    $token = new Token($state, $source, $start, $position);

    // If the state is neither whitespace nor a comment
    if (!$_lexer_whitespace($state) && !$_lexer_comment($state))
    {
        // Set the previous expression and token
        this.expression = $scope && $pop ?
                          $scope.state :
                          '';
        this.token      = $token;
    }

    // Return the current token
    return $token;
}, true);
$_data(__Lexer__, 'peek',   function($position, $language)
{
    // Create a context-free lexer
    var $lexer = new Lexer();

    // Copy the lexer parameters
    $lexer.language = this.language;
    $lexer.position = $position || this.position;
    $lexer.source   = this.source;
    $lexer.state    = $language || this.state;

    // Get the next token
    var $peek = $lexer.next();

    while ($peek)
    {
        // Get the next token type
        var $type = $peek.type;

        // If the next token is neither whitespace nor a comment, break
        if (!$_lexer_whitespace($type) && !$_lexer_comment($type))
            break;

        // Get the next token
        $peek = $lexer.next();
    }

    // Return the next token
    return $peek;
}, true);
$_data(__Lexer__, 'reset',  function()
{
    // Reset the scope chain, expression, position, state, and previous token
    this.chain      = null;
    this.expression = '';
    this.position   = 0;
    this.state      = '';
    this.token      = null;
}, true);